# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["GoDaddyDNSDriver"]

try:
    import simplejson as json
except Exception:
    import json

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import Provider, RecordType, RecordDoesNotExistError
from libcloud.utils.py3 import httplib
from libcloud.common.base import JsonResponse, ConnectionKey
from libcloud.common.types import LibcloudError

API_HOST = "api.godaddy.com"
VALID_RECORD_EXTRA_PARAMS = ["prio", "ttl"]


class GoDaddyDNSException(LibcloudError):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        self.args = (code, message)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<GoDaddyDNSException in {}: {}>".format(self.code, self.message)


class GoDaddyDNSResponse(JsonResponse):
    valid_response_codes = [
        httplib.OK,
        httplib.ACCEPTED,
        httplib.CREATED,
        httplib.NO_CONTENT,
    ]

    def parse_body(self):
        if not self.body:
            return None
        # json.loads doesn't like the regex expressions used in godaddy schema
        self.body = self.body.replace("\\.", "\\\\.")

        data = json.loads(self.body)
        return data

    def parse_error(self):
        data = self.parse_body()
        raise GoDaddyDNSException(code=data["code"], message=data["message"])

    def success(self):
        return self.status in self.valid_response_codes


class GoDaddyDNSConnection(ConnectionKey):
    responseCls = GoDaddyDNSResponse
    host = API_HOST

    allow_insecure = False

    def __init__(
        self,
        key,
        secret,
        secure=True,
        shopper_id=None,
        host=None,
        port=None,
        url=None,
        timeout=None,
        proxy_url=None,
        backoff=None,
        retry_delay=None,
    ):
        super().__init__(
            key,
            secure=secure,
            host=host,
            port=port,
            url=url,
            timeout=timeout,
            proxy_url=proxy_url,
            backoff=backoff,
            retry_delay=retry_delay,
        )
        self.key = key
        self.secret = secret
        self.shopper_id = shopper_id

    def add_default_headers(self, headers):
        if self.shopper_id is not None:
            headers["X-Shopper-Id"] = self.shopper_id
        headers["Content-type"] = "application/json"
        headers["Authorization"] = "sso-key {}:{}".format(self.key, self.secret)
        return headers


class GoDaddyDNSDriver(DNSDriver):
    """
    A driver for GoDaddy DNS.

    This is for customers of GoDaddy
    who wish to purchase, update existing domains
    and manage records for DNS zones owned by GoDaddy NS servers.
    """

    type = Provider.GODADDY
    name = "GoDaddy DNS"
    website = "https://www.godaddy.com/"
    connectionCls = GoDaddyDNSConnection

    RECORD_TYPE_MAP = {
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.MX: "MX",
        RecordType.NS: "SPF",
        RecordType.SRV: "SRV",
        RecordType.TXT: "TXT",
    }

    def __init__(self, shopper_id, key, secret, secure=True, host=None, port=None):
        """
        Instantiate a new `GoDaddyDNSDriver`

        :param  shopper_id: Your customer ID or shopper ID with GoDaddy
        :type   shopper_id: ``str``

        :param  key: Your access key from developer.godaddy.com
        :type   key: ``str``

        :param  secret: Your access key secret
        :type   secret: ``str``
        """
        self.shopper_id = shopper_id
        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            shopper_id=str(shopper_id),
        )

    def list_zones(self):
        """
        Return a list of zones (purchased domains)

        :return: ``list`` of :class:`Zone`
        """
        result = self.connection.request("/v1/domains/").object
        zones = self._to_zones(result)
        return zones

    def list_records(self, zone):
        """
        Return a list of records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :return: ``list`` of :class:`Record`
        """
        result = self.connection.request("/v1/domains/%s/records" % (zone.domain)).object
        records = self._to_records(items=result, zone=zone)
        return records

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a new record.

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param zone: Zone where the requested record is created.
        :type  zone: :class:`Zone`

        :param type: DNS record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: Extra attributes (driver specific). (optional)
        :type extra: ``dict``

        :rtype: :class:`Record`
        """
        new_record = self._format_record(name, type, data, extra)
        self.connection.request(
            "/v1/domains/%s/records" % (zone.domain),
            method="PATCH",
            data=json.dumps([new_record]),
        )
        id = self._get_id_of_record(name, type)
        return Record(
            id=id,
            name=name,
            type=type,
            data=data,
            zone=zone,
            driver=self,
            ttl=new_record["ttl"],
            extra=extra,
        )

    def update_record(self, record, name, type, data, extra=None):
        """
        Update an existing record.

        :param record: Record to update.
        :type  record: :class:`Record`

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param type: DNS record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: (optional) Extra attributes (driver specific).
        :type  extra: ``dict``

        :rtype: :class:`Record`
        """
        new_record = self._format_record(name, type, data, extra)
        self.connection.request(
            "/v1/domains/{}/records/{}/{}".format(record.zone.domain, record.type, record.name),
            method="PUT",
            data=json.dumps([new_record]),
        )
        id = self._get_id_of_record(name, type)
        return Record(
            id=id,
            name=name,
            type=type,
            data=data,
            zone=record.zone,
            driver=self,
            ttl=new_record["ttl"],
            extra=extra,
        )

    def get_record(self, zone_id, record_id):
        """
        Return a Record instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :param record_id: ID of the required record
        :type  record_id: ``str``

        :rtype: :class:`Record`
        """
        parts = record_id.split(":")
        result = self.connection.request(
            "/v1/domains/{}/records/{}/{}".format(zone_id, parts[1], parts[0])
        ).object
        if len(result) == 0:
            raise RecordDoesNotExistError(record_id, driver=self, record_id=record_id)
        return self._to_record(result[0], self.get_zone(zone_id))

    def get_zone(self, zone_id):
        """
        Get a zone (by domain)

        :param  zone_id: The domain, not the ID
        :type   zone_id: ``str``

        :rtype:  :class:`Zone`
        """
        result = self.connection.request("/v1/domains/%s/" % zone_id).object
        zone = self._to_zone(result)
        return zone

    def delete_zone(self, zone):
        """
        Delete a zone.

        Note: This will CANCEL a purchased domain

        :param zone: Zone to delete.
        :type  zone: :class:`Zone`

        :rtype: ``bool``
        """
        self.connection.request("/v1/domains/%s" % (zone.domain), method="DELETE")
        # no error means ok
        return True

    def ex_check_availability(self, domain, for_transfer=False):
        """
        Check the availability of the domain

        :param   domain: the domain name e.g. wazzlewobbleflooble.com
        :type    domain: ``str``

        :param   for_transfer: Check if domain is available for transfer
        :type    for_transfer: ``bool``

        :rtype: `list` of :class:`GoDaddyAvailability`
        """
        result = self.connection.request(
            "/v1/domains/available",
            method="GET",
            params={"domain": domain, "forTransfer": str(for_transfer)},
        ).object
        return GoDaddyAvailability(
            domain=result["domain"],
            available=result["available"],
            price=result["price"],
            currency=result["currency"],
            period=result["period"],
        )

    def ex_list_tlds(self):
        """
        List available TLDs for sale

        :rtype: ``list`` of :class:`GoDaddyTLD`
        """
        result = self.connection.request("/v1/domains/tlds", method="GET").object
        return self._to_tlds(result)

    def ex_get_purchase_schema(self, tld):
        """
        Get the schema that needs completing to purchase a new domain
        Use this in conjunction with ex_purchase_domain

        :param   tld: The top level domain e.g com, eu, uk
        :type    tld: ``str``

        :rtype: `dict` the JSON Schema
        """
        result = self.connection.request(
            "/v1/domains/purchase/schema/%s" % tld, method="GET"
        ).object
        return result

    def ex_get_agreements(self, tld, privacy=True):
        """
        Get the legal agreements for a tld
        Use this in conjunction with ex_purchase_domain

        :param   tld: The top level domain e.g com, eu, uk
        :type    tld: ``str``

        :rtype: `dict` the JSON Schema
        """
        result = self.connection.request(
            "/v1/domains/agreements",
            params={"tlds": tld, "privacy": str(privacy)},
            method="GET",
        ).object
        agreements = []
        for item in result:
            agreements.append(
                GoDaddyLegalAgreement(
                    agreement_key=item["agreementKey"],
                    title=item["title"],
                    url=item["url"],
                    content=item["content"],
                )
            )
        return agreements

    def ex_purchase_domain(self, purchase_request):
        """
        Purchase a domain with GoDaddy

        :param  purchase_request: The completed document
            from ex_get_purchase_schema
        :type   purchase_request: ``dict``

        :rtype: :class:`GoDaddyDomainPurchaseResponse` Your order
        """
        result = self.connection.request(
            "/v1/domains/purchase", data=purchase_request, method="POST"
        ).object
        return GoDaddyDomainPurchaseResponse(
            order_id=result["orderId"],
            item_count=result["itemCount"],
            total=result["total"],
            currency=result["currency"],
        )

    def _format_record(self, name, type, data, extra):
        if extra is None:
            extra = {}
        new_record = {}
        if type == RecordType.SRV:
            new_record = {
                "type": type,
                "name": name,
                "data": data,
                "priority": 1,
                "ttl": extra.get("ttl", 5),
                "service": extra.get("service", ""),
                "protocol": extra.get("protocol", ""),
                "port": extra.get("port", ""),
                "weight": extra.get("weight", "1"),
            }
        else:
            new_record = {
                "type": type,
                "name": name,
                "data": data,
                "ttl": extra.get("ttl", 5),
            }
        if type == RecordType.MX:
            new_record["priority"] = 1
        return new_record

    def _to_zones(self, items):
        zones = []
        for item in items:
            zones.append(self._to_zone(item))
        return zones

    def _to_zone(self, item):
        extra = {"expires": item.get("expires", None)}
        zone = Zone(
            id=item["domainId"],
            domain=item["domain"],
            type="master",
            ttl=None,
            driver=self,
            extra=extra,
        )
        return zone

    def _to_records(self, items, zone=None):
        records = []

        for item in items:
            records.append(self._to_record(item=item, zone=zone))
        return records

    def _to_record(self, item, zone=None):
        ttl = item["ttl"]
        type = self._string_to_record_type(item["type"])
        name = item["name"]
        id = self._get_id_of_record(name, type)
        record = Record(
            id=id,
            name=name,
            type=type,
            data=item["data"],
            zone=zone,
            driver=self,
            ttl=ttl,
        )
        return record

    def _to_tlds(self, items):
        tlds = []
        for item in items:
            tlds.append(self._to_tld(item))
        return tlds

    def _to_tld(self, item):
        return GoDaddyTLD(name=item["name"], tld_type=item["type"])

    def _get_id_of_record(self, name, type):
        return "{}:{}".format(name, type)

    def _ex_connection_class_kwargs(self):
        return {"shopper_id": self.shopper_id}


class GoDaddyAvailability:
    def __init__(self, domain, available, price, currency, period):
        self.domain = domain
        self.available = bool(available)
        # currency comes in micro-units, convert to dollars.
        self.price = float(price) / 1000000
        self.currency = currency
        self.period = int(period)


class GoDaddyTLD:
    def __init__(self, name, tld_type):
        self.name = name
        self.type = tld_type


class GoDaddyDomainPurchaseResponse:
    def __init__(self, order_id, item_count, total, currency):
        self.order_id = order_id
        self.item_count = item_count
        self.total = total
        self.current = currency


class GoDaddyLegalAgreement:
    def __init__(self, agreement_key, title, url, content):
        self.agreement_key = agreement_key
        self.title = title
        self.url = url
        self.content = content



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from libcloud.utils.py3 import httplib
from libcloud.common.onapp import OnAppConnection
from libcloud.compute.base import Node, KeyPair, NodeImage, NodeDriver
from libcloud.utils.networking import is_private_subnet
from libcloud.compute.providers import Provider

__all__ = ["OnAppNodeDriver"]

"""
Define the extra dictionary for specific resources
"""
RESOURCE_EXTRA_ATTRIBUTES_MAP = {
    "node": {
        "add_to_marketplace": {
            "key_name": "add_to_marketplace",
            "transform_func": bool,
        },
        "admin_note": {"key_name": "admin_note", "transform_func": str},
        "allow_resize_without_reboot": {
            "key_name": "allow_resize_without_reboot",
            "transform_func": bool,
        },
        "allowed_hot_migrate": {
            "key_name": "allowed_hot_migrate",
            "transform_func": bool,
        },
        "allowed_swap": {"key_name": "allowed_swap", "transform_func": bool},
        "booted": {"key_name": "booted", "transform_func": bool},
        "built": {"key_name": "built", "transform_func": bool},
        "cpu_priority": {"key_name": "cpu_priority", "transform_func": int},
        "cpu_shares": {"key_name": "cpu_shares", "transform_func": int},
        "cpu_sockets": {"key_name": "cpu_sockets", "transform_func": int},
        "cpu_threads": {"key_name": "cpu_threads", "transform_func": int},
        "cpu_units": {"key_name": "cpu_units", "transform_func": int},
        "cpus": {"key_name": "cpus", "transform_func": int},
        "created_at": {"key_name": "created_at", "transform_func": str},
        "customer_network_id": {
            "key_name": "customer_network_id",
            "transform_func": str,
        },
        "deleted_at": {"key_name": "deleted_at", "transform_func": str},
        "edge_server_type": {"key_name": "edge_server_type", "transform_func": str},
        "enable_autoscale": {"key_name": "enable_autoscale", "transform_func": bool},
        "enable_monitis": {"key_name": "enable_monitis", "transform_func": bool},
        "firewall_notrack": {"key_name": "firewall_notrack", "transform_func": bool},
        "hostname": {"key_name": "hostname", "transform_func": str},
        "hypervisor_id": {"key_name": "hypervisor_id", "transform_func": int},
        "id": {"key_name": "id", "transform_func": int},
        "initial_root_password": {
            "key_name": "initial_root_password",
            "transform_func": str,
        },
        "initial_root_password_encrypted": {
            "key_name": "initial_root_password_encrypted",
            "transform_func": bool,
        },
        "local_remote_access_ip_address": {
            "key_name": "local_remote_access_ip_address",
            "transform_func": str,
        },
        "local_remote_access_port": {
            "key_name": "local_remote_access_port",
            "transform_func": int,
        },
        "locked": {"key_name": "locked", "transform_func": bool},
        "memory": {"key_name": "memory", "transform_func": int},
        "min_disk_size": {"key_name": "min_disk_size", "transform_func": int},
        "monthly_bandwidth_used": {
            "key_name": "monthly_bandwidth_used",
            "transform_func": int,
        },
        "note": {"key_name": "note", "transform_func": str},
        "operating_system": {"key_name": "operating_system", "transform_func": str},
        "operating_system_distro": {
            "key_name": "operating_system_distro",
            "transform_func": str,
        },
        "preferred_hvs": {"key_name": "preferred_hvs", "transform_func": list},
        "price_per_hour": {"key_name": "price_per_hour", "transform_func": float},
        "price_per_hour_powered_off": {
            "key_name": "price_per_hour_powered_off",
            "transform_func": float,
        },
        "recovery_mode": {"key_name": "recovery_mode", "transform_func": bool},
        "remote_access_password": {
            "key_name": "remote_access_password",
            "transform_func": str,
        },
        "service_password": {"key_name": "service_password", "transform_func": str},
        "state": {"key_name": "state", "transform_func": str},
        "storage_server_type": {
            "key_name": "storage_server_type",
            "transform_func": str,
        },
        "strict_virtual_machine_id": {
            "key_name": "strict_virtual_machine_id",
            "transform_func": str,
        },
        "support_incremental_backups": {
            "key_name": "support_incremental_backups",
            "transform_func": bool,
        },
        "suspended": {"key_name": "suspended", "transform_func": bool},
        "template_id": {"key_name": "template_id", "transform_func": int},
        "template_label": {"key_name": "template_label", "transform_func": str},
        "total_disk_size": {"key_name": "total_disk_size", "transform_func": int},
        "updated_at": {"key_name": "updated_at", "transform_func": str},
        "user_id": {"key_name": "user_id", "transform_func": int},
        "vip": {"key_name": "vip", "transform_func": bool},
        "xen_id": {"key_name": "xen_id", "transform_func": int},
    }
}


class OnAppNodeDriver(NodeDriver):
    """
    Base OnApp node driver.
    """

    connectionCls = OnAppConnection
    type = Provider.ONAPP
    name = "OnApp"
    website = "http://onapp.com/"

    def create_node(
        self,
        name,
        ex_memory,
        ex_cpus,
        ex_cpu_shares,
        ex_hostname,
        ex_template_id,
        ex_primary_disk_size,
        ex_swap_disk_size,
        ex_required_virtual_machine_build=1,
        ex_required_ip_address_assignment=1,
        **kwargs,
    ):
        """
        Add a VS

        :param  kwargs: All keyword arguments to create a VS
        :type   kwargs: ``dict``

        :rtype: :class:`OnAppNode`
        """
        server_params = dict(
            label=name,
            memory=ex_memory,
            cpus=ex_cpus,
            cpu_shares=ex_cpu_shares,
            hostname=ex_hostname,
            template_id=ex_template_id,
            primary_disk_size=ex_primary_disk_size,
            swap_disk_size=ex_swap_disk_size,
            required_virtual_machine_build=ex_required_virtual_machine_build,
            required_ip_address_assignment=ex_required_ip_address_assignment,
            rate_limit=kwargs.get("rate_limit"),
        )

        server_params.update(OnAppNodeDriver._create_args_to_params(**kwargs))
        data = json.dumps({"virtual_machine": server_params})

        response = self.connection.request(
            "/virtual_machines.json",
            data=data,
            headers={"Content-type": "application/json"},
            method="POST",
        )

        return self._to_node(response.object["virtual_machine"])

    def destroy_node(self, node, ex_convert_last_backup=0, ex_destroy_all_backups=0):
        """
        Delete a VS

        :param node: OnApp node
        :type  node: :class: `OnAppNode`

        :param convert_last_backup: set 1 to convert the last VS's backup to
                                    template, otherwise set 0
        :type  convert_last_backup: ``int``

        :param destroy_all_backups: set 1 to destroy all existing backups of
                                    this VS, otherwise set 0
        :type  destroy_all_backups: ``int``
        """
        server_params = {
            "convert_last_backup": ex_convert_last_backup,
            "destroy_all_backups": ex_destroy_all_backups,
        }
        action = "/virtual_machines/{identifier}.json".format(identifier=node.id)

        self.connection.request(action, params=server_params, method="DELETE")
        return True

    def list_nodes(self):
        """
        List all VS

        :rtype: ``list`` of :class:`OnAppNode`
        """
        response = self.connection.request("/virtual_machines.json")
        nodes = []
        for vm in response.object:
            nodes.append(self._to_node(vm["virtual_machine"]))
        return nodes

    def list_images(self):
        """
        List all images

        :rtype: ``list`` of :class:`NodeImage`
        """
        response = self.connection.request("/templates.json")
        templates = []
        for template in response.object:
            templates.append(self._to_image(template["image_template"]))
        return templates

    def list_key_pairs(self):
        """
        List all the available key pair objects.

        :rtype: ``list`` of :class:`.KeyPair` objects
        """
        user_id = self.connection.request("/profile.json").object["user"]["id"]
        response = self.connection.request("/users/%s/ssh_keys.json" % user_id)
        ssh_keys = []
        for ssh_key in response.object:
            ssh_keys.append(self._to_key_pair(ssh_key["ssh_key"]))
        return ssh_keys

    def get_key_pair(self, name):
        """
        Retrieve a single key pair.

        :param name: ID of the key pair to retrieve.
        :type name: ``str``

        :rtype: :class:`.KeyPair` object
        """
        user_id = self.connection.request("/profile.json").object["user"]["id"]
        response = self.connection.request("/users/{}/ssh_keys/{}.json".format(user_id, name))
        return self._to_key_pair(response.object["ssh_key"])

    def import_key_pair_from_string(self, name, key_material):
        """
        Import a new public key from string.

        :param name: Key pair name (unused).
        :type name: ``str``

        :param key_material: Public key material.
        :type key_material: ``str``

        :rtype: :class:`.KeyPair` object
        """
        data = json.dumps({"key": key_material})
        user_id = self.connection.request("/profile.json").object["user"]["id"]
        response = self.connection.request(
            "/users/%s/ssh_keys.json" % user_id,
            data=data,
            headers={"Content-type": "application/json"},
            method="POST",
        )
        return self._to_key_pair(response.object["ssh_key"])

    def delete_key_pair(self, key):
        """
        Delete an existing key pair.

        :param key_pair: Key pair object.
        :type key_pair: :class:`.KeyPair`

        :return: True on success
        :rtype: ``bool``
        """
        key_id = key.name
        response = self.connection.request("/settings/ssh_keys/%s.json" % key_id, method="DELETE")
        return response.status == httplib.NO_CONTENT

    #
    # Helper methods
    #

    def _to_key_pair(self, data):
        extra = {"created_at": data["created_at"], "updated_at": data["updated_at"]}
        return KeyPair(
            name=data["id"],
            fingerprint=None,
            public_key=data["key"],
            private_key=None,
            driver=self,
            extra=extra,
        )

    def _to_image(self, template):
        extra = {
            "distribution": template["operating_system_distro"],
            "operating_system": template["operating_system"],
            "operating_system_arch": template["operating_system_arch"],
            "allow_resize_without_reboot": template["allow_resize_without_reboot"],
            "allowed_hot_migrate": template["allowed_hot_migrate"],
            "allowed_swap": template["allowed_swap"],
            "min_disk_size": template["min_disk_size"],
            "min_memory_size": template["min_memory_size"],
            "created_at": template["created_at"],
        }
        return NodeImage(id=template["id"], name=template["label"], driver=self, extra=extra)

    def _to_node(self, data):
        identifier = data["identifier"]
        name = data["label"]
        private_ips = []
        public_ips = []
        for ip in data["ip_addresses"]:
            address = ip["ip_address"]["address"]
            if is_private_subnet(address):
                private_ips.append(address)
            else:
                public_ips.append(address)

        extra = OnAppNodeDriver._get_extra_dict(data, RESOURCE_EXTRA_ATTRIBUTES_MAP["node"])
        return Node(identifier, name, extra["state"], public_ips, private_ips, self, extra=extra)

    @staticmethod
    def _get_extra_dict(response, mapping):
        """
        Extract attributes from the element based on rules provided in the
        mapping dictionary.

        :param   response: The JSON response to parse the values from.
        :type    response: ``dict``

        :param   mapping: Dictionary with the extra layout
        :type    mapping: ``dict``

        :rtype:  ``dict``
        """
        extra = {}
        for attribute, values in mapping.items():
            transform_func = values["transform_func"]
            value = response.get(values["key_name"])

            extra[attribute] = transform_func(value) if value else None
        return extra

    @staticmethod
    def _create_args_to_params(**kwargs):
        """
        Extract server params from keyword args to create a VS

        :param   kwargs: keyword args
        :return: ``dict``
        """
        params = [
            "ex_cpu_sockets",
            "ex_cpu_threads",
            "ex_enable_autoscale",
            "ex_data_store_group_primary_id",
            "ex_data_store_group_swap_id",
            "ex_hypervisor_group_id",
            "ex_hypervisor_id",
            "ex_initial_root_password",
            "ex_note",
            "ex_primary_disk_min_iops",
            "ex_primary_network_id",
            "ex_primary_network_group_id",
            "ex_recipe_ids",
            "ex_required_automatic_backup",
            "ex_required_virtual_machine_startup",
            "ex_required_virtual_machine_startup",
            "ex_selected_ip_address_id",
            "ex_swap_disk_min_iops",
            "ex_type_of_format",
            "ex_custom_recipe_variables",
            "ex_licensing_key",
            "ex_licensing_server_id",
            "ex_licensing_type",
        ]
        server_params = {}

        for p in params:
            value = kwargs.get(p)
            if value:
                server_params[p[3:]] = value
        return server_params



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import simplejson as json
except ImportError:
    import json  # NOQA

from libcloud.loadbalancer.base import Driver, Member, Algorithm, LoadBalancer
from libcloud.compute.drivers.gce import GCEConnection, GCENodeDriver

# GCE doesn't actually give you an algorithm choice, but this is here simply as
# the closest match.  The actual algorithm is described here:
# https://developers.google.com/compute/docs/load-balancing/#overview
DEFAULT_ALGORITHM = Algorithm.RANDOM


class GCELBDriver(Driver):
    connectionCls = GCEConnection
    apiname = "googleapis"
    name = "Google Compute Engine Load Balancer"
    website = "https://cloud.google.com/"

    _VALUE_TO_ALGORITHM_MAP = {"RANDOM": Algorithm.RANDOM}

    def __init__(self, *args, **kwargs):
        if kwargs.get("gce_driver"):
            self.gce = kwargs["gce_driver"]
        else:
            self.gce = GCENodeDriver(*args, **kwargs)

        self.connection = self.gce.connection

    def _get_node_from_ip(self, ip):
        """
        Return the node object that matches a given public IP address.

        :param  ip: Public IP address to search for
        :type   ip: ``str``

        :return:  Node object that has the given IP, or None if not found.
        :rtype:   :class:`Node` or None
        """
        all_nodes = self.gce.list_nodes(ex_zone="all")
        for node in all_nodes:
            if ip in node.public_ips:
                return node
        return None

    def list_protocols(self):
        """
        Return a list of supported protocols.

        For GCE, this is simply a hardcoded list.

        :rtype: ``list`` of ``str``
        """
        return ["TCP", "UDP"]

    def list_balancers(self, ex_region=None):
        """
        List all loadbalancers

        :keyword  ex_region: The region to return balancers from.  If None,
                             will default to self.region.  If 'all', will
                             return all balancers.
        :type     ex_region: ``str`` or :class:`GCERegion` or ``None``

        :rtype: ``list`` of :class:`LoadBalancer`
        """
        balancers = []
        for fwr in self.gce.ex_list_forwarding_rules(region=ex_region):
            balancers.append(self._forwarding_rule_to_loadbalancer(fwr))
        return balancers

    def create_balancer(
        self,
        name,
        port,
        protocol,
        algorithm,
        members,
        ex_region=None,
        ex_healthchecks=None,
        ex_address=None,
        ex_session_affinity=None,
    ):
        """
        Create a new load balancer instance.

        For GCE, this means creating a forwarding rule and a matching target
        pool, then adding the members to the target pool.

        :param  name: Name of the new load balancer (required)
        :type   name: ``str``

        :param  port: Port or range of ports the load balancer should listen
                      on, defaults to all ports.  Examples: '80', '5000-5999'
        :type   port: ``str``

        :param  protocol: Load balancer protocol.  Should be 'tcp' or 'udp',
                          defaults to 'tcp'.
        :type   protocol: ``str``

        :param  members: List of Members to attach to balancer.  Can be Member
                         objects or Node objects.  Node objects are preferred
                         for GCE, but Member objects are accepted to comply
                         with the established libcloud API.  Note that the
                         'port' attribute of the members is ignored.
        :type   members: ``list`` of :class:`Member` or :class:`Node`

        :param  algorithm: Load balancing algorithm.  Ignored for GCE which
                           uses a hashing-based algorithm.
        :type   algorithm: :class:`Algorithm` or ``None``

        :keyword  ex_region:  Optional region to create the load balancer in.
                              Defaults to the default region of the GCE Node
                              Driver.
        :type     ex_region:  C{GCERegion} or ``str``

        :keyword  ex_healthchecks: Optional list of healthcheck objects or
                                   names to add to the load balancer.
        :type     ex_healthchecks: ``list`` of :class:`GCEHealthCheck` or
                                   ``list`` of ``str``

        :keyword  ex_address: Optional static address object to be assigned to
                              the load balancer.
        :type     ex_address: C{GCEAddress}

        :keyword  ex_session_affinity: Optional algorithm to use for session
                                       affinity.  This will modify the hashing
                                       algorithm such that a client will tend
                                       to stick to a particular Member.
        :type     ex_session_affinity: ``str``

        :return:  LoadBalancer object
        :rtype:   :class:`LoadBalancer`
        """
        node_list = []
        for member in members:
            # Member object
            if hasattr(member, "ip"):
                if member.extra.get("node"):
                    node_list.append(member.extra["node"])
                else:
                    node_list.append(self._get_node_from_ip(member.ip))
            # Node object
            elif hasattr(member, "name"):
                node_list.append(member)
            # Assume it's a node name otherwise
            else:
                node_list.append(self.gce.ex_get_node(member, "all"))

        # Create Target Pool
        tp_name = "%s-tp" % name
        targetpool = self.gce.ex_create_targetpool(
            tp_name,
            region=ex_region,
            healthchecks=ex_healthchecks,
            nodes=node_list,
            session_affinity=ex_session_affinity,
        )

        # Create the Forwarding rule, but if it fails, delete the target pool.
        try:
            forwarding_rule = self.gce.ex_create_forwarding_rule(
                name,
                targetpool,
                region=ex_region,
                protocol=protocol,
                port_range=port,
                address=ex_address,
            )
        except Exception:
            targetpool.destroy()
            raise

        # Reformat forwarding rule to LoadBalancer object
        return self._forwarding_rule_to_loadbalancer(forwarding_rule)

    def destroy_balancer(self, balancer):
        """
        Destroy a load balancer.

        For GCE, this means destroying the associated forwarding rule, then
        destroying the target pool that was attached to the forwarding rule.

        :param  balancer: LoadBalancer which should be used
        :type   balancer: :class:`LoadBalancer`

        :return:  True if successful
        :rtype:   ``bool``
        """
        destroy = balancer.extra["forwarding_rule"].destroy()
        if destroy:
            tp_destroy = balancer.extra["targetpool"].destroy()
            return tp_destroy
        else:
            return destroy

    def get_balancer(self, balancer_id):
        """
        Return a :class:`LoadBalancer` object.

        :param  balancer_id: Name of load balancer you wish to fetch.  For GCE,
                             this is the name of the associated forwarding
                             rule.
        :param  balancer_id: ``str``

        :rtype: :class:`LoadBalancer`
        """
        fwr = self.gce.ex_get_forwarding_rule(balancer_id)
        return self._forwarding_rule_to_loadbalancer(fwr)

    def balancer_attach_compute_node(self, balancer, node):
        """
        Attach a compute node as a member to the load balancer.

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param node: Node to join to the balancer
        :type  node: :class:`Node`

        :return: Member after joining the balancer.
        :rtype: :class:`Member`
        """
        add_node = balancer.extra["targetpool"].add_node(node)
        if add_node:
            return self._node_to_member(node, balancer)

    def balancer_attach_member(self, balancer, member):
        """
        Attach a member to balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param member: Member to join to the balancer
        :type member: :class:`Member`

        :return: Member after joining the balancer.
        :rtype: :class:`Member`
        """
        node = member.extra.get("node") or self._get_node_from_ip(member.ip)
        add_node = balancer.extra["targetpool"].add_node(node)
        if add_node:
            return self._node_to_member(node, balancer)

    def balancer_detach_member(self, balancer, member):
        """
        Detach member from balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param member: Member which should be used
        :type member: :class:`Member`

        :return: True if member detach was successful, otherwise False
        :rtype: ``bool``
        """
        node = member.extra.get("node") or self._get_node_from_ip(member.ip)
        remove_node = balancer.extra["targetpool"].remove_node(node)
        return remove_node

    def balancer_list_members(self, balancer):
        """
        Return list of members attached to balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :rtype: ``list`` of :class:`Member`
        """
        return [self._node_to_member(n, balancer) for n in balancer.extra["targetpool"].nodes]

    def ex_create_healthcheck(self, *args, **kwargs):
        return self.gce.ex_create_healthcheck(*args, **kwargs)

    def ex_list_healthchecks(self):
        return self.gce.ex_list_healthchecks()

    def ex_balancer_attach_healthcheck(self, balancer, healthcheck):
        """
        Attach a healthcheck to balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param healthcheck: Healthcheck to add
        :type  healthcheck: :class:`GCEHealthCheck`

        :return: True if successful
        :rtype:  ``bool``
        """
        return balancer.extra["targetpool"].add_healthcheck(healthcheck)

    def ex_balancer_detach_healthcheck(self, balancer, healthcheck):
        """
        Detach healtcheck from balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param healthcheck: Healthcheck to remove
        :type  healthcheck: :class:`GCEHealthCheck`

        :return: True if successful
        :rtype: ``bool``
        """
        return balancer.extra["targetpool"].remove_healthcheck(healthcheck)

    def ex_balancer_list_healthchecks(self, balancer):
        """
        Return list of healthchecks attached to balancer

        :param  balancer: LoadBalancer which should be used
        :type   balancer: :class:`LoadBalancer`

        :rtype: ``list`` of :class:`HealthChecks`
        """
        return balancer.extra["healthchecks"]

    def _node_to_member(self, node, balancer):
        """
        Return a Member object based on a Node.

        :param  node: Node object
        :type   node: :class:`Node`

        :keyword  balancer: The balancer the member is attached to.
        :type     balancer: :class:`LoadBalancer`

        :return:  Member object
        :rtype:   :class:`Member`
        """
        # A balancer can have a node as a member, even if the node doesn't
        # exist.  In this case, 'node' is simply a string to where the resource
        # would be found if it was there.
        if hasattr(node, "name"):
            member_id = node.name
        else:
            member_id = node

        if hasattr(node, "public_ips") and len(node.public_ips) > 0:
            member_ip = node.public_ips[0]
        else:
            member_ip = None

        extra = {"node": node}
        return Member(
            id=member_id,
            ip=member_ip,
            port=balancer.port,
            balancer=balancer,
            extra=extra,
        )

    def _forwarding_rule_to_loadbalancer(self, forwarding_rule):
        """
        Return a Load Balancer object based on a GCEForwardingRule object.

        :param  forwarding_rule: ForwardingRule object
        :type   forwarding_rule: :class:`GCEForwardingRule`

        :return:  LoadBalancer object
        :rtype:   :class:`LoadBalancer`
        """
        extra = {}
        extra["forwarding_rule"] = forwarding_rule
        extra["targetpool"] = forwarding_rule.targetpool
        extra["healthchecks"] = forwarding_rule.targetpool.healthchecks

        return LoadBalancer(
            id=forwarding_rule.id,
            name=forwarding_rule.name,
            state=None,
            ip=forwarding_rule.address,
            port=forwarding_rule.extra["portRange"],
            driver=self,
            extra=extra,
        )
